"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.PossiblyNormalCompletion = exports.JoinedAbruptCompletions = exports.ReturnCompletion = exports.BreakCompletion = exports.ContinueCompletion = exports.IntrospectionThrowCompletion = exports.ThrowCompletion = exports.NormalCompletion = exports.AbruptCompletion = exports.Completion = undefined;

var _index = require("./values/index.js");

var _invariant = require("./invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } } /**
                                                                                                                                                           * Copyright (c) 2017-present, Facebook, Inc.
                                                                                                                                                           * All rights reserved.
                                                                                                                                                           *
                                                                                                                                                           * This source code is licensed under the BSD-style license found in the
                                                                                                                                                           * LICENSE file in the root directory of this source tree. An additional grant
                                                                                                                                                           * of patent rights can be found in the PATENTS file in the same directory.
                                                                                                                                                           */

var Completion = exports.Completion = function Completion(value, target) {
  _classCallCheck(this, Completion);

  this.value = value;
  this.target = target;
};
// Abrupt completions are thrown as exeptions, to make it a easier
// to quickly get to the matching high level construct.


var AbruptCompletion = exports.AbruptCompletion = function (_Completion) {
  _inherits(AbruptCompletion, _Completion);

  function AbruptCompletion() {
    _classCallCheck(this, AbruptCompletion);

    return _possibleConstructorReturn(this, (AbruptCompletion.__proto__ || Object.getPrototypeOf(AbruptCompletion)).apply(this, arguments));
  }

  return AbruptCompletion;
}(Completion);
// Normal completions are returned just like spec completions


var NormalCompletion = exports.NormalCompletion = function (_Completion2) {
  _inherits(NormalCompletion, _Completion2);

  function NormalCompletion() {
    _classCallCheck(this, NormalCompletion);

    return _possibleConstructorReturn(this, (NormalCompletion.__proto__ || Object.getPrototypeOf(NormalCompletion)).apply(this, arguments));
  }

  return NormalCompletion;
}(Completion);

var ThrowCompletion = exports.ThrowCompletion = function (_AbruptCompletion) {
  _inherits(ThrowCompletion, _AbruptCompletion);

  function ThrowCompletion(value, nativeStack) {
    _classCallCheck(this, ThrowCompletion);

    var _this3 = _possibleConstructorReturn(this, (ThrowCompletion.__proto__ || Object.getPrototypeOf(ThrowCompletion)).call(this, value));

    (0, _invariant2.default)(value.getType() !== value.$Realm.intrinsics.__IntrospectionError || _this3 instanceof IntrospectionThrowCompletion);
    _this3.nativeStack = nativeStack || new Error().stack;
    return _this3;
  }

  return ThrowCompletion;
}(AbruptCompletion);

var IntrospectionThrowCompletion = exports.IntrospectionThrowCompletion = function (_ThrowCompletion) {
  _inherits(IntrospectionThrowCompletion, _ThrowCompletion);

  function IntrospectionThrowCompletion() {
    _classCallCheck(this, IntrospectionThrowCompletion);

    return _possibleConstructorReturn(this, (IntrospectionThrowCompletion.__proto__ || Object.getPrototypeOf(IntrospectionThrowCompletion)).apply(this, arguments));
  }

  return IntrospectionThrowCompletion;
}(ThrowCompletion);

var ContinueCompletion = exports.ContinueCompletion = function (_AbruptCompletion2) {
  _inherits(ContinueCompletion, _AbruptCompletion2);

  function ContinueCompletion() {
    _classCallCheck(this, ContinueCompletion);

    return _possibleConstructorReturn(this, (ContinueCompletion.__proto__ || Object.getPrototypeOf(ContinueCompletion)).apply(this, arguments));
  }

  return ContinueCompletion;
}(AbruptCompletion);

var BreakCompletion = exports.BreakCompletion = function (_AbruptCompletion3) {
  _inherits(BreakCompletion, _AbruptCompletion3);

  function BreakCompletion() {
    _classCallCheck(this, BreakCompletion);

    return _possibleConstructorReturn(this, (BreakCompletion.__proto__ || Object.getPrototypeOf(BreakCompletion)).apply(this, arguments));
  }

  return BreakCompletion;
}(AbruptCompletion);

var ReturnCompletion = exports.ReturnCompletion = function (_AbruptCompletion4) {
  _inherits(ReturnCompletion, _AbruptCompletion4);

  function ReturnCompletion() {
    _classCallCheck(this, ReturnCompletion);

    return _possibleConstructorReturn(this, (ReturnCompletion.__proto__ || Object.getPrototypeOf(ReturnCompletion)).apply(this, arguments));
  }

  return ReturnCompletion;
}(AbruptCompletion);

var JoinedAbruptCompletions = exports.JoinedAbruptCompletions = function (_AbruptCompletion5) {
  _inherits(JoinedAbruptCompletions, _AbruptCompletion5);

  function JoinedAbruptCompletions(realm, joinCondition, consequent, consequentEffects, alternate, alternateEffects) {
    _classCallCheck(this, JoinedAbruptCompletions);

    var _this8 = _possibleConstructorReturn(this, (JoinedAbruptCompletions.__proto__ || Object.getPrototypeOf(JoinedAbruptCompletions)).call(this, realm.intrinsics.empty, undefined));

    _this8.joinCondition = joinCondition;
    _this8.consequent = consequent;
    _this8.consequentEffects = consequentEffects;
    _this8.alternate = alternate;
    _this8.alternateEffects = alternateEffects;
    return _this8;
  }

  return JoinedAbruptCompletions;
}(AbruptCompletion);

// Possibly normal completions have to be treated like normal completions
// and are thus never thrown. At the end of a try block or loop body, however,
// action must be taken to deal with the possibly abrupt case of the completion.


var PossiblyNormalCompletion = exports.PossiblyNormalCompletion = function (_NormalCompletion) {
  _inherits(PossiblyNormalCompletion, _NormalCompletion);

  function PossiblyNormalCompletion(value, joinCondition, consequent, consequentEffects, alternate, alternateEffects) {
    _classCallCheck(this, PossiblyNormalCompletion);

    (0, _invariant2.default)(consequent instanceof NormalCompletion || consequent instanceof _index.Value || alternate instanceof NormalCompletion || alternate instanceof _index.Value);
    (0, _invariant2.default)(consequent instanceof AbruptCompletion || alternate instanceof AbruptCompletion);
    (0, _invariant2.default)(value === consequent || consequent instanceof AbruptCompletion || consequent instanceof NormalCompletion && value === consequent.value);
    (0, _invariant2.default)(value === alternate || alternate instanceof AbruptCompletion || alternate instanceof NormalCompletion && value === alternate.value);

    var _this9 = _possibleConstructorReturn(this, (PossiblyNormalCompletion.__proto__ || Object.getPrototypeOf(PossiblyNormalCompletion)).call(this, value));

    _this9.joinCondition = joinCondition;
    _this9.consequent = consequent;
    _this9.consequentEffects = consequentEffects;
    _this9.alternate = alternate;
    _this9.alternateEffects = alternateEffects;
    return _this9;
  }

  return PossiblyNormalCompletion;
}(NormalCompletion);
//# sourceMappingURL=completions.js.map