"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (realm) {
  // ECMA262 22.1.3.30
  return new _index.NativeFunctionValue(realm, "Object.prototype.toString", "toString", 0, function (context) {
    // 1. If the this value is undefined, return "[object Undefined]".
    if (context instanceof _index.UndefinedValue) return new _index.StringValue(realm, "[object Undefined]");

    // 2. If the this value is null, return "[object Null]".
    if (context instanceof _index.NullValue) return new _index.StringValue(realm, "[object Null]");

    // 3. Let O be ToObject(this value).
    var O = (0, _to.ToObjectPartial)(realm, context);

    var builtinTag = void 0;

    // 4. Let isArray be ? IsArray(O).
    var isArray = (0, _is.IsArray)(realm, O);

    // 5. If isArray is true, let builtinTag be "Array".
    if (isArray) builtinTag = "Array";

    // 6. Else, if O is an exotic String object, let builtinTag be "String".
    else if (O.$StringData) builtinTag = "String";

      // 7. Else, if O has an [[ParameterMap]] internal slot, let builtinTag be "Arguments".
      else if (O.$ParameterMap) builtinTag = "Arguments";

        // 8. Else, if O has a [[Call]] internal method, let builtinTag be "Function".
        else if (O.$Call) builtinTag = "Function";

          // 9. Else, if O has an [[ErrorData]] internal slot, let builtinTag be "Error".
          else if (O.$ErrorData) builtinTag = "Error";

            // 10. Else, if O has a [[BooleanData]] internal slot, let builtinTag be "Boolean".
            else if (O.$BooleanData) builtinTag = "Boolean";

              // 11. Else, if O has a [[NumberData]] internal slot, let builtinTag be "Number".
              else if (O.$NumberData) builtinTag = "Number";

                // 12. Else, if O has a [[DateValue]] internal slot, let builtinTag be "Date".
                else if (O.$DateValue) builtinTag = "Date";

                  // 13. Else, if O has a [[RegExpMatcher]] internal slot, let builtinTag be "RegExp".
                  else if (O.$RegExpMatcher) builtinTag = "RegExp";

                    // 14. Else, let builtinTag be "Object".
                    else builtinTag = "Object";

    // 15. Let tag be ? Get(O, @@toStringTag).
    var tag = (0, _get.Get)(realm, O, realm.intrinsics.SymbolToStringTag);

    // 16. If Type(tag) is not String, let tag be builtinTag.
    tag = tag instanceof _index.StringValue ? tag.value : builtinTag;

    // 17. Return the String that is the result of concatenating "[object ", tag, and "]".
    return new _index.StringValue(realm, "[object " + tag + "]");
  }, false);
};

var _index = require("../../values/index.js");

var _to = require("../../methods/to.js");

var _is = require("../../methods/is.js");

var _get = require("../../methods/get.js");
//# sourceMappingURL=ObjectProto_toString.js.map