"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _prepackStandalone = require("./prepack-standalone");

Object.keys(_prepackStandalone).forEach(function (key) {
  if (key === "default" || key === "__esModule") return;
  Object.defineProperty(exports, key, {
    enumerable: true,
    get: function get() {
      return _prepackStandalone[key];
    }
  });
});

var _prepackNodeEnvironment = require("./prepack-node-environment");

Object.keys(_prepackNodeEnvironment).forEach(function (key) {
  if (key === "default" || key === "__esModule") return;
  Object.defineProperty(exports, key, {
    enumerable: true,
    get: function get() {
      return _prepackNodeEnvironment[key];
    }
  });
});
exports.prepackString = prepackString;
exports.prepackStdin = prepackStdin;
exports.prepackFile = prepackFile;
exports.prepackFileSync = prepackFileSync;

var _index = require("./serializer/index.js");

var _index2 = _interopRequireDefault(_index);

var _construct_realm = require("./construct_realm.js");

var _construct_realm2 = _interopRequireDefault(_construct_realm);

var _globals = require("./globals.js");

var _globals2 = _interopRequireDefault(_globals);

var _fs = require("fs");

var _fs2 = _interopRequireDefault(_fs);

var _completions = require("./completions.js");

var _options = require("./options");

var _errors = require("./errors.js");

var _invariant = require("./invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Copyright (c) 2017-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

function prepackString(filename, code, sourceMap) {
  var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : _options.defaultOptions;

  var realm = (0, _construct_realm2.default)((0, _options.getRealmOptions)(options));
  (0, _globals2.default)(realm);
  if (options.serialize || !options.residual) {
    var serializer = new _index2.default(realm, (0, _options.getSerializerOptions)(options));
    var serialized = serializer.init(options.filename || filename, code, sourceMap, options.sourceMaps);
    if (!serialized) {
      throw _errors.fatalError;
    }
    if (!options.residual) return serialized;
    var result = realm.$GlobalEnv.executePartialEvaluator(filename, serialized.code, JSON.stringify(serialized.map));
    if (result instanceof _completions.AbruptCompletion) throw result;
    return result;
  } else {
    (0, _invariant2.default)(options.residual);
    var _result = realm.$GlobalEnv.executePartialEvaluator(filename, code, sourceMap);
    if (_result instanceof _completions.AbruptCompletion) throw _result;
    return _result;
  }
}

function prepackStdin() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : _options.defaultOptions;
  var callback = arguments[1];

  var sourceMapFilename = options.inputSourceMapFilename || '';
  process.stdin.setEncoding('utf8');
  process.stdin.resume();
  process.stdin.on('data', function (code) {
    _fs2.default.readFile(sourceMapFilename, "utf8", function (mapErr, sourceMap) {
      if (mapErr) {
        console.warn("No sourcemap found at " + sourceMapFilename + ".");
        sourceMap = "";
      }
      var filename = 'no-filename-specified';
      var serialized = void 0;
      try {
        serialized = prepackString(filename, code, sourceMap, options);
      } catch (err) {
        callback(err, null);
        return;
      }
      callback(null, serialized);
    });
  });
}

function prepackFile(filename) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : _options.defaultOptions;
  var callback = arguments[2];
  var errorHandler = arguments[3];

  if (options.compatibility === 'node-cli') {
    (0, _prepackNodeEnvironment.prepackNodeCLI)(filename, options, callback);
    return;
  }
  var sourceMapFilename = options.inputSourceMapFilename || filename + ".map";
  _fs2.default.readFile(filename, "utf8", function (fileErr, code) {
    if (fileErr) {
      if (errorHandler) errorHandler(fileErr);
      return;
    }
    _fs2.default.readFile(sourceMapFilename, "utf8", function (mapErr, sourceMap) {
      if (mapErr) {
        console.warn("No sourcemap found at " + sourceMapFilename + ".");
        sourceMap = "";
      }
      var serialized = void 0;
      try {
        serialized = prepackString(filename, code, sourceMap, options);
      } catch (err) {
        callback(err, null);
        return;
      }
      callback(null, serialized);
    });
  });
}

function prepackFileSync(filename) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : _options.defaultOptions;

  if (options.compatibility === 'node-cli') {
    return (0, _prepackNodeEnvironment.prepackNodeCLISync)(filename, options);
  }
  var code = _fs2.default.readFileSync(filename, "utf8");
  var sourceMap = "";
  var sourceMapFilename = options.inputSourceMapFilename || filename + ".map";
  try {
    sourceMap = _fs2.default.readFileSync(sourceMapFilename, "utf8");
  } catch (_e) {
    console.warn("No sourcemap found at " + sourceMapFilename + ".");
  }
  return prepackString(filename, code, sourceMap, options);
}
//# sourceMappingURL=prepack-node.js.map